# -*- coding=utf-8 -*-

"""Module for the RunMetadata class."""

from collections.abc import Iterable
from pathlib import Path
from typing import Optional

from pulp import LpStatus, LpStatusOptimal
from revsymg.index_lib import IndexT
from yaml import dump  # type: ignore

from khloraascaf.cli import (
    ARG_CONTIG_ATTRS,
    ARG_CONTIG_LINKS,
    ARG_CONTIG_STARTER,
    OPT_INSTANCE_NAME,
    OPT_MULT_UPB,
    OPT_OUTDEBUG,
    OPT_OUTDIR,
    OPT_PRESSCORE_UPB,
    OPT_SOLVER,
)
from khloraascaf.inputs import STR_ORIENT, IdCT, MultT, PresScoreT
from khloraascaf.lib import (
    DR_REGION_ID,
    IR_REGION_ID,
    REGION_CODE_TO_ID,
    RegionCodeT,
)
from khloraascaf.multiplied_doubled_contig_graph import (
    CIND_IND,
    COCC_IND,
    COR_IND,
    OccOrCT,
)
from khloraascaf.outputs import (
    CONTIGS_OF_REGIONS_PREFIX,
    MAP_OF_REGIONS_PREFIX,
    ORIENT_INT_STR,
    fmt_contigs_of_regions_filename,
    fmt_map_of_regions_filename,
)
from khloraascaf.utils_debug import (
    FOUND_REPFRAG_PREFIX,
    VERTICES_OF_REGIONS_PREFIX,
    fmt_found_repeated_fragments_filename,
    fmt_vertices_of_regions_filename,
)


# DOCU metadata run YAML all
# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     Files                                    #
# ---------------------------------------------------------------------------- #
# DOCU run metadata file
YAML_EXT = 'yaml'
"""Extension for YAML files"""

IO_CONFIG_METADATA_PREFIX = 'io_config'
"""Prefix of the IO config metadata file name."""

SOLUTIONS_METADATA_PREFIX = 'solutions'
"""Prefix of the solutions metadata file name."""

DEBUG_METADATA_PREFIX = 'all_scaffolding'
"""Prefix of the all scaffolding metadata file name."""

# ---------------------------------------------------------------------------- #
#                                     YAML                                     #
# ---------------------------------------------------------------------------- #
KEY_ILP_COMBINATION = 'ilp_combination'
KEY_STARTER_VERTEX = 'starter_vertex'
KEY_OPT_VALUE = 'opt_value'
KEY_ILP_STATUS = 'ilp_status'


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                               String Formatter                               #
# ---------------------------------------------------------------------------- #
def vertex_to_str(vertex: OccOrCT) -> str:
    """Transform a vertex to a string.

    Parameters
    ----------
    vertex : OccOrCT
        Vertex

    Returns
    -------
    str
        String representation
    """
    return (
        f'{vertex[CIND_IND]}'
        f'\t{ORIENT_INT_STR[vertex[COR_IND]]}'
        f'\t{vertex[COCC_IND]}'
    )


def str_to_vertex(vertex_str: str) -> OccOrCT:
    """Transform a vertex string to the vertex.

    Parameters
    ----------
    vertex_str : str
        Vertex string

    Returns
    -------
    OccOrCT
        Vertex
    """
    str_list = vertex_str.split('\t')
    return (
        IndexT(str_list[CIND_IND]),
        STR_ORIENT[str_list[COR_IND]],  # type: ignore
        IndexT(str_list[COCC_IND]),
    )


# ---------------------------------------------------------------------------- #
#                                 YAML Outputs                                 #
# ---------------------------------------------------------------------------- #
# pylint: disable=too-many-arguments
def bump_io_config_to_yaml(output_dir: Path, contig_attrs: Path,
                           contig_links: Path, contig_starter: IdCT,
                           mult_upb: MultT, presscore_upb: PresScoreT,
                           solver: str, outdir: Path, instance_name: str,
                           debug: bool):
    """Add input metadata.

    Parameters
    ----------
    output_dir : Path
        YAML file output directory path
    contig_attrs : Path
        Contig attributes file path
    contig_links : Path
        Contig links file path
    contig_starter : IdCT
        Starter contig's identifier
    mult_upb : MultT
        Multiplicities upper bound
    presscore_upb : PresScoreT
        Presence score upper bound
    solver : str
        MILP solver to use
    outdir : Path
        Output directory path
    instance_name : str
        Instance name
    debug : bool
        Output debug or not
    """
    yaml_path = output_dir / fmt_io_config_metadata_filename()
    with open(yaml_path, 'a', encoding='utf-8') as yaml_out:
        yaml_out.write(
            dump(
                {
                    ARG_CONTIG_ATTRS: str(contig_attrs),
                    ARG_CONTIG_LINKS: str(contig_links),
                    ARG_CONTIG_STARTER: contig_starter,
                    OPT_MULT_UPB: mult_upb,
                    OPT_PRESSCORE_UPB: presscore_upb,
                    OPT_SOLVER: solver,
                    OPT_OUTDIR: str(outdir),
                    OPT_INSTANCE_NAME: instance_name,
                    OPT_OUTDEBUG: debug,
                },
                sort_keys=False,
            ),
        )


def fmt_io_config_metadata_filename() -> str:
    """Return IO config metadata file name.

    Returns
    -------
    str
        File name
    """
    return f'{IO_CONFIG_METADATA_PREFIX}.{YAML_EXT}'


def bump_solutions_to_yaml(output_dir: Path, instance_name: str,
                           ilp_codes: Iterable[RegionCodeT]):
    """Append debug metadata to the YAML file.

    Parameters
    ----------
    output_dir : Path
        YAML file output directory path
    instance_name : str
        Instance name
    ilp_codes : iterable of RegionCodeT
        ILP codes
    """
    yaml_path = output_dir / fmt_solutions_metadata_filename()

    ilp_ids = [REGION_CODE_TO_ID[code] for code in ilp_codes]
    with open(yaml_path, 'a', encoding='utf-8') as yaml_out:
        yaml_out.write(
            dump([
                {
                    KEY_ILP_COMBINATION: ilp_ids,
                    CONTIGS_OF_REGIONS_PREFIX: fmt_contigs_of_regions_filename(
                        instance_name, ilp_ids,
                    ),
                    MAP_OF_REGIONS_PREFIX: fmt_map_of_regions_filename(
                        instance_name, ilp_ids,
                    ),
                },
            ],
                sort_keys=False,
            ),
        )


def fmt_solutions_metadata_filename() -> str:
    """Return solutionsg metadata file name.

    Returns
    -------
    str
        File name
    """
    return f'{SOLUTIONS_METADATA_PREFIX}.{YAML_EXT}'


# pylint: disable=too-many-arguments
def bump_debug_to_yaml(output_dir: Path, instance_name: str,
                       ilp_codes: Iterable[RegionCodeT],
                       starter_vertex: OccOrCT, status: int,
                       opt_value: Optional[float]):
    """Append debug metadata to the YAML file.

    Parameters
    ----------
    output_dir : Path
        YAML file output directory path
    instance_name : str
        Instance name
    ilp_codes : iterable of RegionCodeT
        ILP codes
    starter_vertex : OccOrCT
        Starter vertex
    status : int
        ILP status code
    opt_value : optional float
        ILP optimal value if exists else None
    """
    ilp_ids = [REGION_CODE_TO_ID[code] for code in ilp_codes]
    dict_debug = {
        KEY_ILP_COMBINATION: ilp_ids,
        KEY_STARTER_VERTEX: vertex_to_str(starter_vertex),
        KEY_ILP_STATUS: LpStatus[status],
    }
    if status == LpStatusOptimal:
        assert opt_value is not None
        dict_debug.update(
            {
                KEY_OPT_VALUE: opt_value,
                VERTICES_OF_REGIONS_PREFIX: fmt_vertices_of_regions_filename(
                    instance_name, ilp_ids,
                ),
                MAP_OF_REGIONS_PREFIX: fmt_map_of_regions_filename(
                    instance_name, ilp_ids,
                ),
            },
        )
        if ilp_ids[-1] in (IR_REGION_ID, DR_REGION_ID):
            dict_debug.update(
                {
                    FOUND_REPFRAG_PREFIX: fmt_found_repeated_fragments_filename(
                        instance_name, ilp_ids,
                    ),
                },
            )

    yaml_path = output_dir / fmt_debug_metadata_filename()
    with open(yaml_path, 'a', encoding='utf-8') as yaml_out:
        yaml_out.write(dump([dict_debug], sort_keys=False))


def fmt_debug_metadata_filename() -> str:
    """Return debug metadata file name.

    Returns
    -------
    str
        File name
    """
    return f'{DEBUG_METADATA_PREFIX}.{YAML_EXT}'



# -*- coding=utf-8 -*-

"""Functional testing with toy examples.

See tests/data/README.md
"""

# pylint: disable=compare-to-zero,missing-raises-doc
import subprocess
from pathlib import Path

from khloraascaf.inputs import INSTANCE_NAME_DEF, SOLVER_CBC
from khloraascaf.lib import DR_REGION_ID, IR_REGION_ID, SC_REGION_ID
from khloraascaf.outputs import (
    fmt_contigs_of_regions_filename,
    fmt_map_of_regions_filename,
)
from khloraascaf.scaffolding_methods import scaffolding
from tests.path_utils import rm


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
TEST_DIR: Path = Path(__file__).parent.absolute()

_TOY_DATADIR = TEST_DIR / 'data'

# ---------------------------------------------------------------------------- #
#                                    IR - SC                                   #
# ---------------------------------------------------------------------------- #
_IR_SC_DIR = _TOY_DATADIR / 'ir_sc'
_IR_SC_CONTIG_ATTRS = _IR_SC_DIR / 'contig_attrs.tsv'
_IR_SC_CONTIG_LINKS = _IR_SC_DIR / 'contig_links.tsv'
_IR_SC_CONTIG_STARTER = 'C0'
_IR_SC_SOL_REGMAP = _IR_SC_DIR / 'map_of_regions_sol.tsv'
_IR_SC_SOL_REGCTG_F = _IR_SC_DIR / 'contigs_of_regions_sol_0.tsv'
_IR_SC_SOL_REGCTG_R = _IR_SC_DIR / 'contigs_of_regions_sol_1.tsv'

# ---------------------------------------------------------------------------- #
#                                    DR - SC                                   #
# ---------------------------------------------------------------------------- #
_DR_SC_DIR = _TOY_DATADIR / 'dr_sc'
_DR_SC_CONTIG_ATTRS = _DR_SC_DIR / 'contig_attrs.tsv'
_DR_SC_CONTIG_LINKS = _DR_SC_DIR / 'contig_links.tsv'
_DR_SC_CONTIG_STARTER = 'C0'
_DR_SC_SOL_REGMAP = _DR_SC_DIR / 'map_of_regions_sol.tsv'
_DR_SC_SOL_REGCTG = _DR_SC_DIR / 'contigs_of_regions_sol.tsv'

# ---------------------------------------------------------------------------- #
#                                      SC                                      #
# ---------------------------------------------------------------------------- #
_SC_DIR = _TOY_DATADIR / 'sc'
_SC_CONTIG_ATTRS = _SC_DIR / 'contig_attrs.tsv'
_SC_CONTIG_LINKS = _SC_DIR / 'contig_links.tsv'
_SC_CONTIG_STARTER = 'C0'
_SC_SOL_REGMAP = _SC_DIR / 'map_of_regions_sol.tsv'
_SC_SOL_REGCTG = _SC_DIR / 'contigs_of_regions_sol.tsv'

# ---------------------------------------------------------------------------- #
#                                 IR - DR - SC                                 #
# ---------------------------------------------------------------------------- #
_IR_DR_SC_DIR = _TOY_DATADIR / 'ir_dr_sc'
_IR_DR_SC_CONTIG_ATTRS = _IR_DR_SC_DIR / 'contig_attrs.tsv'
_IR_DR_SC_CONTIG_LINKS = _IR_DR_SC_DIR / 'contig_links.tsv'
_IR_DR_SC_CONTIG_STARTER = 'C0'
_IR_DR_SC_SOL_REGMAP = _IR_DR_SC_DIR / 'map_of_regions_sol.tsv'
_IR_DR_SC_SOL_REGCTG_F = _IR_DR_SC_DIR / 'contigs_of_regions_sol_0.tsv'
_IR_DR_SC_SOL_REGCTG_R = _IR_DR_SC_DIR / 'contigs_of_regions_sol_1.tsv'

# ---------------------------------------------------------------------------- #
#                                 DR - IR - SC                                 #
# ---------------------------------------------------------------------------- #
_DR_IR_SC_DIR = _TOY_DATADIR / 'dr_ir_sc'
_DR_IR_SC_CONTIG_ATTRS = _DR_IR_SC_DIR / 'contig_attrs.tsv'
_DR_IR_SC_CONTIG_LINKS = _DR_IR_SC_DIR / 'contig_links.tsv'
_DR_IR_SC_CONTIG_STARTER = 'C0'
_DR_IR_SC_SOL_REGMAP = _DR_IR_SC_DIR / 'map_of_regions_sol.tsv'
_DR_IR_SC_SOL_REGCTG_F = _DR_IR_SC_DIR / 'contigs_of_regions_sol_0.tsv'
_DR_IR_SC_SOL_REGCTG_R = _DR_IR_SC_DIR / 'contigs_of_regions_sol_1.tsv'


# ============================================================================ #
#                                TEST FUNCTIONS                                #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                    IR - SC                                   #
# ---------------------------------------------------------------------------- #
def test_ir_sc_func():
    """Test IR-SC toy example."""
    outdir = _IR_SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    outdir_gen = scaffolding(
        _IR_SC_CONTIG_ATTRS,
        _IR_SC_CONTIG_LINKS,
        _IR_SC_CONTIG_STARTER,
        solver=SOLVER_CBC,
        outdir=outdir,
        instance_name=INSTANCE_NAME_DEF,
        debug=True,
    )
    # TOTEST verify all the debug file
    verify_scaffolding_ir_sc(outdir_gen)
    rm(outdir)


def test_ir_sc_cli():
    """Test IR-SC toy example for cli."""
    outdir = _IR_SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    try:
        proc = subprocess.run(
            [
                'python3',
                '-m',
                'khloraascaf',
                _IR_SC_CONTIG_ATTRS,
                _IR_SC_CONTIG_LINKS,
                _IR_SC_CONTIG_STARTER,
                '--solver',
                SOLVER_CBC,
                '--debug',
                '--out-directory',
                outdir,
            ],
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as err:
        raise AssertionError() from err
    outdir_gen = Path(proc.stdout.splitlines()[-1])
    verify_scaffolding_ir_sc(outdir_gen)
    rm(outdir)


def verify_scaffolding_ir_sc(outdir_gen: Path):
    """Verify scaffolding IR-SC.

    Parameters
    ----------
    outdir_gen : Path
        Output directory
    """
    #
    # Test output files
    #
    assert {p.name for p in outdir_gen.glob('*')} == {
        'all_scaffolding.yaml',
        'contigs_of_regions_khloraascaf_ir_sc.tsv',
        'io_config.yaml',
        'map_of_regions_khloraascaf_dr.tsv',
        'map_of_regions_khloraascaf_ir_dr.tsv',
        'map_of_regions_khloraascaf_ir_sc.tsv',
        'map_of_regions_khloraascaf_ir.tsv',
        'repfrag_khloraascaf_dr.tsv',
        'repfrag_khloraascaf_ir_dr.tsv',
        'repfrag_khloraascaf_ir.tsv',
        'solutions.yaml',
        'solver_cbc_khloraascaf_dr.log',
        'solver_cbc_khloraascaf_ir_dr.log',
        'solver_cbc_khloraascaf_ir_sc.log',
        'solver_cbc_khloraascaf_ir.log',
        'vertices_of_regions_khloraascaf_dr.tsv',
        'vertices_of_regions_khloraascaf_ir_dr.tsv',
        'vertices_of_regions_khloraascaf_ir_sc.tsv',
        'vertices_of_regions_khloraascaf_ir.tsv',
    }
    #
    # Test maps of regions
    #
    res_map_of_regions = outdir_gen / fmt_map_of_regions_filename(
        INSTANCE_NAME_DEF, (IR_REGION_ID, SC_REGION_ID),
    )
    l_sol_map = []
    with open(_IR_SC_SOL_REGMAP, 'r', encoding='utf-8') as sol_map:
        for line in sol_map:
            l_sol_map.append(line.split())
    l_res_map = []
    with open(res_map_of_regions, 'r', encoding='utf-8') as res_map:
        for line in res_map:
            l_res_map.append(line.split())
    assert l_sol_map == l_res_map
    #
    # Test contigs of region
    #
    res_contigs_of_regions = outdir_gen / fmt_contigs_of_regions_filename(
        INSTANCE_NAME_DEF, (IR_REGION_ID, SC_REGION_ID),
    )
    l_sol_ctg_f = []
    with open(_IR_SC_SOL_REGCTG_F, 'r', encoding='utf-8') as sol_ctg:
        for line in sol_ctg:
            l_sol_ctg_f.append(line.split())
    l_sol_ctg_r = []
    with open(_IR_SC_SOL_REGCTG_R, 'r', encoding='utf-8') as sol_ctg:
        for line in sol_ctg:
            l_sol_ctg_r.append(line.split())
    l_res_ctg = []
    with open(res_contigs_of_regions, 'r', encoding='utf-8') as res_ctg:
        for line in res_ctg:
            l_res_ctg.append(line.split())
    assert l_res_ctg in (l_sol_ctg_f, l_sol_ctg_r)


# ---------------------------------------------------------------------------- #
#                                    DR - SC                                   #
# ---------------------------------------------------------------------------- #
def test_dr_sc_func():
    """Test DR-SC toy example."""
    outdir = _DR_SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    outdir_gen = scaffolding(
        _DR_SC_CONTIG_ATTRS,
        _DR_SC_CONTIG_LINKS,
        _DR_SC_CONTIG_STARTER,
        solver=SOLVER_CBC,
        debug=True,
        outdir=outdir,
    )
    verify_scaffolding_dr_sc(outdir_gen)
    rm(outdir)


def test_dr_sc_cli():
    """Test DR-SC toy example for cli."""
    outdir = _DR_SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    try:
        proc = subprocess.run(
            [
                'python3',
                '-m',
                'khloraascaf',
                _DR_SC_CONTIG_ATTRS,
                _DR_SC_CONTIG_LINKS,
                _DR_SC_CONTIG_STARTER,
                '--solver',
                SOLVER_CBC,
                '--debug',
                '--out-directory',
                outdir,
            ],
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as err:
        raise AssertionError() from err
    outdir_gen = Path(proc.stdout.splitlines()[-1])
    verify_scaffolding_dr_sc(outdir_gen)
    rm(outdir)


def verify_scaffolding_dr_sc(outdir_gen: Path):
    """Verify scaffolding DR-SC.

    Parameters
    ----------
    outdir_gen : Path
        Output directory
    """
    #
    # Test output files
    #
    assert {p.name for p in outdir_gen.glob('*')} == {
        'all_scaffolding.yaml',
        'contigs_of_regions_khloraascaf_dr_sc.tsv',
        'io_config.yaml',
        'map_of_regions_khloraascaf_dr_ir.tsv',
        'map_of_regions_khloraascaf_dr_sc.tsv',
        'map_of_regions_khloraascaf_dr.tsv',
        'map_of_regions_khloraascaf_ir.tsv',
        'repfrag_khloraascaf_dr_ir.tsv',
        'repfrag_khloraascaf_dr.tsv',
        'repfrag_khloraascaf_ir.tsv',
        'solutions.yaml',
        'solver_cbc_khloraascaf_dr_ir.log',
        'solver_cbc_khloraascaf_dr.log',
        'solver_cbc_khloraascaf_ir.log',
        'solver_cbc_khloraascaf_dr_sc.log',
        'vertices_of_regions_khloraascaf_dr_ir.tsv',
        'vertices_of_regions_khloraascaf_dr.tsv',
        'vertices_of_regions_khloraascaf_ir.tsv',
        'vertices_of_regions_khloraascaf_dr_sc.tsv',
    }
    #
    # Test map of regions
    #
    res_map_of_regions = outdir_gen / fmt_map_of_regions_filename(
        INSTANCE_NAME_DEF, (DR_REGION_ID, SC_REGION_ID),
    )
    l_sol_map = []
    with open(_DR_SC_SOL_REGMAP, 'r', encoding='utf-8') as sol_map:
        for line in sol_map:
            l_sol_map.append(line.split())
    l_res_map = []
    with open(res_map_of_regions, 'r', encoding='utf-8') as res_map:
        for line in res_map:
            l_res_map.append(line.split())
    assert l_sol_map == l_res_map
    #
    # Test contigs of regions
    #
    res_contigs_of_regions = outdir_gen / fmt_contigs_of_regions_filename(
        INSTANCE_NAME_DEF, (DR_REGION_ID, SC_REGION_ID),
    )
    l_sol_ctg = []
    with open(_DR_SC_SOL_REGCTG, 'r', encoding='utf-8') as sol_ctg:
        for line in sol_ctg:
            l_sol_ctg.append(line.split())
    l_res_ctg = []
    with open(res_contigs_of_regions, 'r', encoding='utf-8') as res_ctg:
        for line in res_ctg:
            l_res_ctg.append(line.split())
    assert l_res_ctg == l_sol_ctg


# ---------------------------------------------------------------------------- #
#                                      SC                                      #
# ---------------------------------------------------------------------------- #
def test_sc_func():
    """Test SC toy example."""
    outdir = _SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    outdir_gen = scaffolding(
        _SC_CONTIG_ATTRS,
        _SC_CONTIG_LINKS,
        _SC_CONTIG_STARTER,
        solver=SOLVER_CBC,
        debug=True,
        outdir=outdir,
    )
    verify_scaffolding_sc(outdir_gen)
    rm(outdir)


def test_sc_cli():
    """Test SC toy example for cli."""
    outdir = _SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    try:
        proc = subprocess.run(
            [
                'python3',
                '-m',
                'khloraascaf',
                _SC_CONTIG_ATTRS,
                _SC_CONTIG_LINKS,
                _SC_CONTIG_STARTER,
                '--solver',
                SOLVER_CBC,
                '--debug',
                '--out-directory',
                outdir,
            ],
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as err:
        raise AssertionError() from err
    outdir_gen = Path(proc.stdout.splitlines()[-1])
    verify_scaffolding_sc(outdir_gen)
    rm(outdir)


def verify_scaffolding_sc(outdir_gen: Path):
    """Verify scaffolding SC.

    Parameters
    ----------
    outdir_gen : Path
        Output directory
    """
    #
    # Test output files
    #
    assert {p.name for p in outdir_gen.glob('*')} == {
        'all_scaffolding.yaml',
        'contigs_of_regions_khloraascaf_sc.tsv',
        'io_config.yaml',
        'map_of_regions_khloraascaf_dr.tsv',
        'map_of_regions_khloraascaf_ir.tsv',
        'map_of_regions_khloraascaf_sc.tsv',
        'repfrag_khloraascaf_dr.tsv',
        'repfrag_khloraascaf_ir.tsv',
        'solutions.yaml',
        'solver_cbc_khloraascaf_dr.log',
        'solver_cbc_khloraascaf_ir.log',
        'solver_cbc_khloraascaf_sc.log',
        'vertices_of_regions_khloraascaf_dr.tsv',
        'vertices_of_regions_khloraascaf_ir.tsv',
        'vertices_of_regions_khloraascaf_sc.tsv',
    }
    #
    # Test map of regions
    #
    res_map_of_regions = outdir_gen / fmt_map_of_regions_filename(
        INSTANCE_NAME_DEF, (SC_REGION_ID,),
    )
    l_sol_map = []
    with open(_SC_SOL_REGMAP, 'r', encoding='utf-8') as sol_map:
        for line in sol_map:
            l_sol_map.append(line.split())
    l_res_map = []
    with open(res_map_of_regions, 'r', encoding='utf-8') as res_map:
        for line in res_map:
            l_res_map.append(line.split())
    assert l_sol_map == l_res_map
    #
    # Test contigs of regions
    #
    res_contigs_of_regions = outdir_gen / fmt_contigs_of_regions_filename(
        INSTANCE_NAME_DEF, (SC_REGION_ID,),
    )
    l_sol_ctg = []
    with open(_SC_SOL_REGCTG, 'r', encoding='utf-8') as sol_ctg:
        for line in sol_ctg:
            l_sol_ctg.append(line.split())
    l_res_ctg = []
    with open(res_contigs_of_regions, 'r', encoding='utf-8') as res_ctg:
        for line in res_ctg:
            l_res_ctg.append(line.split())
    assert l_res_ctg == l_sol_ctg


# ---------------------------------------------------------------------------- #
#                                 IR - DR - SC                                 #
# ---------------------------------------------------------------------------- #
def test_ir_dr_sc_func():
    """Test IR-DR-SC toy example."""
    outdir = _IR_DR_SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    outdir_gen = scaffolding(
        _IR_DR_SC_CONTIG_ATTRS,
        _IR_DR_SC_CONTIG_LINKS,
        _IR_DR_SC_CONTIG_STARTER,
        solver=SOLVER_CBC,
        outdir=outdir,
        instance_name=INSTANCE_NAME_DEF,
        debug=True,
    )
    # TOTEST verify all the debug file
    verify_scaffolding_ir_dr_sc(outdir_gen)
    rm(outdir)


def test_ir_dr_sc_cli():
    """Test IR-DR-SC alone toy example for cli."""
    outdir = _IR_DR_SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    try:
        proc = subprocess.run(
            [
                'python3',
                '-m',
                'khloraascaf',
                _IR_DR_SC_CONTIG_ATTRS,
                _IR_DR_SC_CONTIG_LINKS,
                _IR_DR_SC_CONTIG_STARTER,
                '--solver',
                SOLVER_CBC,
                '--debug',
                '--out-directory',
                outdir,
            ],
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as err:
        raise AssertionError() from err
    outdir_gen = Path(proc.stdout.splitlines()[-1])
    verify_scaffolding_ir_dr_sc(outdir_gen)
    rm(outdir)


def verify_scaffolding_ir_dr_sc(outdir_gen: Path):
    """Verify scaffolding IR-DR-SC.

    Parameters
    ----------
    outdir_gen : Path
        Output directory
    """
    #
    # Test output files
    #
    assert {p.name for p in outdir_gen.glob('*')} == {
        'all_scaffolding.yaml',
        'contigs_of_regions_khloraascaf_ir_dr_sc.tsv',
        'io_config.yaml',
        'map_of_regions_khloraascaf_dr.tsv',
        'map_of_regions_khloraascaf_ir_dr_sc.tsv',
        'map_of_regions_khloraascaf_ir_dr.tsv',
        'map_of_regions_khloraascaf_ir.tsv',
        'repfrag_khloraascaf_dr.tsv',
        'repfrag_khloraascaf_ir_dr.tsv',
        'repfrag_khloraascaf_ir.tsv',
        'solutions.yaml',
        'solver_cbc_khloraascaf_dr.log',
        'solver_cbc_khloraascaf_ir_dr_sc.log',
        'solver_cbc_khloraascaf_ir_dr.log',
        'solver_cbc_khloraascaf_ir.log',
        'vertices_of_regions_khloraascaf_dr.tsv',
        'vertices_of_regions_khloraascaf_ir_dr_sc.tsv',
        'vertices_of_regions_khloraascaf_ir_dr.tsv',
        'vertices_of_regions_khloraascaf_ir.tsv',
    }
    #
    # Test maps of regions
    #
    res_map_of_regions = outdir_gen / fmt_map_of_regions_filename(
        INSTANCE_NAME_DEF, (IR_REGION_ID, DR_REGION_ID, SC_REGION_ID),
    )
    l_sol_map = []
    with open(_IR_DR_SC_SOL_REGMAP, 'r', encoding='utf-8') as sol_map:
        for line in sol_map:
            l_sol_map.append(line.split())
    l_res_map = []
    with open(res_map_of_regions, 'r', encoding='utf-8') as res_map:
        for line in res_map:
            l_res_map.append(line.split())
    assert l_sol_map == l_res_map
    #
    # Test contigs of region
    #
    res_contigs_of_regions = outdir_gen / fmt_contigs_of_regions_filename(
        INSTANCE_NAME_DEF, (IR_REGION_ID, DR_REGION_ID, SC_REGION_ID),
    )
    l_sol_ctg_f = []
    with open(_IR_DR_SC_SOL_REGCTG_F, 'r', encoding='utf-8') as sol_ctg:
        for line in sol_ctg:
            l_sol_ctg_f.append(line.split())
    l_sol_ctg_r = []
    with open(_IR_DR_SC_SOL_REGCTG_R, 'r', encoding='utf-8') as sol_ctg:
        for line in sol_ctg:
            l_sol_ctg_r.append(line.split())
    l_res_ctg = []
    with open(res_contigs_of_regions, 'r', encoding='utf-8') as res_ctg:
        for line in res_ctg:
            l_res_ctg.append(line.split())
    assert l_res_ctg in (l_sol_ctg_f, l_sol_ctg_r)


# ---------------------------------------------------------------------------- #
#                                 DR - IR - SC                                 #
# ---------------------------------------------------------------------------- #
def test_dr_ir_sc_func():
    """Test DR-IR-SC toy example."""
    outdir = _DR_IR_SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    outdir_gen = scaffolding(
        _DR_IR_SC_CONTIG_ATTRS,
        _DR_IR_SC_CONTIG_LINKS,
        _DR_IR_SC_CONTIG_STARTER,
        solver=SOLVER_CBC,
        outdir=outdir,
        instance_name=INSTANCE_NAME_DEF,
        debug=True,
    )
    # TOTEST verify all the debug file
    verify_scaffolding_dr_ir_sc(outdir_gen)
    rm(outdir)


def test_dr_ir_sc_cli():
    """Test DR-IR-SC alone toy example for cli."""
    outdir = _DR_IR_SC_DIR / 'tmp'
    outdir.mkdir(exist_ok=True)
    try:
        proc = subprocess.run(
            [
                'python3',
                '-m',
                'khloraascaf',
                _DR_IR_SC_CONTIG_ATTRS,
                _DR_IR_SC_CONTIG_LINKS,
                _DR_IR_SC_CONTIG_STARTER,
                '--solver',
                SOLVER_CBC,
                '--debug',
                '--out-directory',
                outdir,
            ],
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as err:
        raise AssertionError() from err
    outdir_gen = Path(proc.stdout.splitlines()[-1])
    verify_scaffolding_dr_ir_sc(outdir_gen)
    rm(outdir)


def verify_scaffolding_dr_ir_sc(outdir_gen: Path):
    """Verify scaffolding DR-IR-SC.

    Parameters
    ----------
    outdir_gen : Path
        Output directory
    """
    #
    # Test output files
    #
    assert {p.name for p in outdir_gen.glob('*')} == {
        'all_scaffolding.yaml',
        'contigs_of_regions_khloraascaf_dr_ir_sc.tsv',
        'io_config.yaml',
        'map_of_regions_khloraascaf_dr_ir_sc.tsv',
        'map_of_regions_khloraascaf_dr_ir.tsv',
        'map_of_regions_khloraascaf_dr.tsv',
        'map_of_regions_khloraascaf_ir.tsv',
        'repfrag_khloraascaf_dr_ir.tsv',
        'repfrag_khloraascaf_dr.tsv',
        'repfrag_khloraascaf_ir.tsv',
        'solutions.yaml',
        'solver_cbc_khloraascaf_dr_ir_sc.log',
        'solver_cbc_khloraascaf_dr_ir.log',
        'solver_cbc_khloraascaf_dr.log',
        'solver_cbc_khloraascaf_ir.log',
        'vertices_of_regions_khloraascaf_dr_ir_sc.tsv',
        'vertices_of_regions_khloraascaf_dr_ir.tsv',
        'vertices_of_regions_khloraascaf_dr.tsv',
        'vertices_of_regions_khloraascaf_ir.tsv',
    }
    #
    # Test maps of regions
    #
    res_map_of_regions = outdir_gen / fmt_map_of_regions_filename(
        INSTANCE_NAME_DEF, (DR_REGION_ID, IR_REGION_ID, SC_REGION_ID),
    )
    l_sol_map = []
    with open(_DR_IR_SC_SOL_REGMAP, 'r', encoding='utf-8') as sol_map:
        for line in sol_map:
            l_sol_map.append(line.split())
    l_res_map = []
    with open(res_map_of_regions, 'r', encoding='utf-8') as res_map:
        for line in res_map:
            l_res_map.append(line.split())
    assert l_sol_map == l_res_map
    #
    # Test contigs of region
    #
    res_contigs_of_regions = outdir_gen / fmt_contigs_of_regions_filename(
        INSTANCE_NAME_DEF, (DR_REGION_ID, IR_REGION_ID, SC_REGION_ID),
    )
    l_sol_ctg_f = []
    with open(_DR_IR_SC_SOL_REGCTG_F, 'r', encoding='utf-8') as sol_ctg:
        for line in sol_ctg:
            l_sol_ctg_f.append(line.split())
    l_sol_ctg_r = []
    with open(_DR_IR_SC_SOL_REGCTG_R, 'r', encoding='utf-8') as sol_ctg:
        for line in sol_ctg:
            l_sol_ctg_r.append(line.split())
    l_res_ctg = []
    with open(res_contigs_of_regions, 'r', encoding='utf-8') as res_ctg:
        for line in res_ctg:
            l_res_ctg.append(line.split())
    assert l_res_ctg in (l_sol_ctg_f, l_sol_ctg_r)



# -*- coding=utf-8 -*-

"""Caffolding result's module: structures and utils."""


from __future__ import annotations

from queue import LifoQueue, Queue
from typing import Iterable, Iterator, Optional

from pulp import LpProblem
from revsymg.index_lib import FORWARD_INT, REVERSE_INT, IndexT, OrT

from khloraascaf.exceptions import NotACircuit
from khloraascaf.ilp.dirf_sets import dirf_canonical, dirf_other
from khloraascaf.ilp.invf_sets import invf_canonical, invf_other
from khloraascaf.ilp.pulp_var_db import (
    BIN_THRESHOLD,
    PuLPVarDirFModel,
    PuLPVarInvFModel,
    PuLPVarModelT,
)
from khloraascaf.lib import DR_CODE, IR_CODE, SC_CODE, RegionCodeT
from khloraascaf.multiplied_doubled_contig_graph import (
    COCC_IND,
    COR_IND,
    MDCGraph,
    OccOrCT,
)


# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                Genomic Regions                               #
# ---------------------------------------------------------------------------- #
OrRegT = tuple[IndexT, OrT]
OrRegMapT = tuple[OrRegT, ...]
RegOccOrCT = tuple[OccOrCT, ...]
#
# Builder
#
_OrRegMapT = list[OrRegT]
_RegOccOrCT = list[OccOrCT]


# ============================================================================ #
#                                     CLASS                                    #
# ============================================================================ #
# pylint: disable=too-many-instance-attributes
class ScaffoldingResult():
    """Scaffolding result class.

    An instance contains a genomic regions map and each genomic regions'
    oriented contigs.

    Warning
    -------
    You must not call the constructor, you must only deal with the objects
    that have been already instanced.
    You can use the class name to type your variables.
    """

    # DOCU update docstring
    # DOCU add to docstring example to get regions' orc from result

    # pylint: disable=too-many-arguments
    def __init__(self, ilp_codes: tuple[RegionCodeT, ...],
                 status: int, opt_value: float,
                 region_map: OrRegMapT,
                 regions_occorc: tuple[RegOccOrCT, ...],
                 sc_regions: tuple[IndexT, ...],
                 ir_regions: tuple[IndexT, ...],
                 dr_regions: tuple[IndexT, ...]):
        """The Initializer."""
        #
        # Ilp origin
        #
        self.__ilp_codes: tuple[RegionCodeT, ...] = ilp_codes
        #
        # ILP result
        #
        self.__status: int = status
        self.__opt_value: float = opt_value
        #
        # Genomic region
        #
        self.__region_map: OrRegMapT = region_map
        self.__regions_occorc: tuple[RegOccOrCT, ...] = regions_occorc
        self.__sc_regions: tuple[IndexT, ...] = sc_regions
        self.__ir_regions: tuple[IndexT, ...] = ir_regions
        self.__dr_regions: tuple[IndexT, ...] = dr_regions

    # ~*~ Getter ~*~

    def last_ilp(self) -> RegionCodeT:
        """Return last ILP code.

        Returns
        -------
        RegionCodeT
            Last ILP code
        """
        return self.__ilp_codes[-1]

    def ilp_codes(self) -> Iterator[RegionCodeT]:
        """Iterate over ILP codes.

        Yields
        ------
        RegionCodeT
            ILP code
        """
        yield from self.__ilp_codes

    def status(self) -> int:
        """Returns status code.

        Returns
        -------
        int
            Status code
        """
        return self.__status

    def opt_value(self) -> float:
        """ILP optimal value.

        Returns
        -------
        float
            ILP optimal value
        """
        return self.__opt_value

    def map_of_regions(self) -> Iterator[tuple[IndexT, OrT]]:
        """Iterate over the regions and their orientation.

        Yields
        ------
        IndexT
            Region's index
        OrT
            Region's orientation
        """
        yield from self.__region_map

    def region_occorc(self, region_index: IndexT) -> Iterator[OccOrCT]:
        """Iterate over the multiplied oriented contigs of the region.

        Parameters
        ----------
        region_index : IndexT
            Region's index

        Yields
        ------
        OccOrCT
            Multiplied oriented contig of the region
        """
        yield from self.__regions_occorc[region_index]

    def number_regions(self) -> int:
        """Returns the number of regions.

        Returns
        -------
        int
            Number of regions
        """
        return len(self.__regions_occorc)

    def sc_regions(self) -> Iterator[IndexT]:
        """Iterate over the single copy indices.

        Yields
        ------
        IndexT
            Single copy region index
        """
        yield from self.__sc_regions

    def ir_regions(self) -> Iterator[IndexT]:
        """Iterate over inverted repeat indices.

        Yields
        ------
        IndexT
            Inverted repeats index
        """
        yield from self.__ir_regions

    def dr_regions(self) -> Iterator[IndexT]:
        """Iterate over direct repeat indices.

        Yields
        ------
        IndexT
            Direct repeats index
        """
        yield from self.__dr_regions


class _ScaffoldingResultBuilder():
    """The scaffolding result builder."""

    __NO_OPT_VALUE = .0

    @classmethod
    def __fix_opt_value(cls, prob: LpProblem) -> float:
        """Fix the #XXX PuLP issue 331.

        Parameters
        ----------
        prob : LpProblem
            PuLP problem

        Returns
        -------
        float
            Optimal value
        """
        if prob.objective.value() is not None:
            return prob.objective.value()
        return cls.__NO_OPT_VALUE

    def __init__(self):
        """The Initializer."""
        #
        # Genomic region
        #
        self.__region_map: _OrRegMapT = []
        self.__regions_occorc: list[_RegOccOrCT] = []
        self.__sc_regions: list[IndexT] = []
        self.__ir_regions: list[IndexT] = []
        self.__dr_regions: list[IndexT] = []

    # ~*~ Setter ~*~

    def add_sc_region(self) -> IndexT:
        """Add an single copy region.

        Returns
        -------
        IndexT
            Region's index
        """
        region_index = len(self.__regions_occorc)
        self.__regions_occorc.append([])
        self.__region_map.append((region_index, FORWARD_INT))
        self.__sc_regions.append(region_index)
        return region_index

    def add_ir_region(self,
                      region_index: Optional[IndexT] = None) -> IndexT:
        """Add an inverted repeat region.

         #XXX region_index provides the first repeat
         * this works because we build pair of repeated regions

        Parameters
        ----------
        region_index : IndexT, optional
            Index of an already existing region, else None

        Returns
        -------
        IndexT
            Region's index
        """
        #
        # Non-existing region
        #
        if region_index is None:
            region_index = len(self.__regions_occorc)
            self.__regions_occorc.append([])
            self.__region_map.append((region_index, FORWARD_INT))
            self.__ir_regions.append(region_index)
        #
        # Already existing region (repeat)
        #
        else:
            self.__region_map.append((region_index, REVERSE_INT))
        return region_index

    def add_dr_region(self,
                      region_index: Optional[IndexT] = None) -> IndexT:
        """Add a direct repeat region.

         #XXX region_index provides the first repeat
         * this works because we build pair of repeated regions

        Parameters
        ----------
        region_index : IndexT, default None
            Index of an already existing region, else None

        Returns
        -------
        IndexT
            Region's index
        """
        #
        # Non-existing region
        #
        if region_index is None:
            region_index = len(self.__regions_occorc)
            self.__regions_occorc.append([])
            self.__region_map.append((region_index, FORWARD_INT))
            self.__dr_regions.append(region_index)
        #
        # Already existing region (repeat)
        #
        else:
            self.__region_map.append((region_index, FORWARD_INT))
        return region_index

    def add_occorc_to_region(self, v: OccOrCT, region_index: IndexT):
        """Add v to the region denoted by its index.

        Parameters
        ----------
        v : OccOrCT
            Multiplied oriented contig
        region_index : IndexT
            Region's index
        """
        self.__regions_occorc[region_index].append(v)

    # ~*~ Getter ~*~

    def view(self, prob: LpProblem,
             ilp_codes: Iterable[RegionCodeT]) -> ScaffoldingResult:
        """Return a ScaffoldingResult view from the builder.

        Parameters
        ----------
        prob : LpProblem
            PuLP problem
        ilp_codes : iterable of RegionCodeT
            ILP codes

        Returns
        -------
        ScaffoldingResult
            Scaffolding result view
        """
        return ScaffoldingResult(
            tuple(ilp_codes),
            prob.status,
            self.__fix_opt_value(prob),
            tuple(self.__region_map),
            tuple(tuple(regoccorc) for regoccorc in self.__regions_occorc),
            tuple(self.__sc_regions),
            tuple(self.__ir_regions),
            tuple(self.__dr_regions),
        )


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                               Build The Regions                              #
# ---------------------------------------------------------------------------- #
# pylint: disable=too-many-arguments
def path_to_regions(mdcg: MDCGraph, starter_vertex: OccOrCT,
                    ilp_codes: Iterable[RegionCodeT],
                    prob: LpProblem, var: PuLPVarModelT,
                    fix_result: Optional[ScaffoldingResult] = None) -> (
                        ScaffoldingResult):
    """Extract regions from optimal path.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    ilp_codes : iterable of RegionCodeT
        Code of the regions that have been scaffolded
    prob : LpProblem
        PuLP problem
    var : PuLPVarModelT
        PuLP variables
    fix_result : ScaffoldingResult, optional
        Scaffolding result with regions, by default `None`

    Returns
    -------
    ScaffoldingResult
        Scaffolding result
    """
    # REFACTOR should path_to_regions be a method for result builder?
    # ------------------------------------------------------------------------ #
    # Manage paired fragments
    # ------------------------------------------------------------------------ #
    set_invf_paired = __create_set_invf_paired(var, fix_result)
    set_dirf_paired = __create_set_dirf_paired(var, fix_result)

    # ------------------------------------------------------------------------ #
    # Init regions
    # ------------------------------------------------------------------------ #
    result_builder = _ScaffoldingResultBuilder()
    reg_code = SC_CODE
    region_index = result_builder.add_sc_region()

    initial_vertex = __find_initial(
        mdcg, starter_vertex, var, set_invf_paired, set_dirf_paired)

    u: OccOrCT = initial_vertex  # previous v
    v: Optional[OccOrCT] = initial_vertex

    #
    # Region IR: LIFO, DR: FIFO
    #
    # XXX no order between pairs of repeats
    #   * for the moment, the pairs of contiguous repeat are not considered
    #       ordered (e.g. pairs of contiguous IR should be lifo, DR fifo)
    ir_lifo: dict[IndexT, LifoQueue[OccOrCT]] = {}
    dr_fifo: dict[IndexT, Queue[OccOrCT]] = {}
    #
    # Canonical to the region's index: repeat was discovered
    #
    ir_canonical_reg: dict[OccOrCT, IndexT] = {}
    dr_canonical_reg: dict[OccOrCT, IndexT] = {}

    # ------------------------------------------------------------------------ #
    # Walk into the solution path
    # ------------------------------------------------------------------------ #
    while v is not None:
        new_reg_code = __get_region_code(v, set_invf_paired, set_dirf_paired)
        #
        # Type is contiguous
        #
        if new_reg_code == reg_code:
            #
            # Previous UNIQ region
            #
            if new_reg_code == SC_CODE:  # pylint: disable=compare-to-zero
                result_builder.add_occorc_to_region(v, region_index)
            #
            # Previous or new IR region
            #
            elif new_reg_code == IR_CODE:
                region_index = __previous_or_new_ir(
                    new_reg_code, region_index, u, v,
                    var, ir_lifo, ir_canonical_reg, result_builder,
                )
            #
            # Previous or new DR region
            #
            elif new_reg_code == DR_CODE:
                __previous_or_new_dr(
                    new_reg_code, region_index, u, v,
                    var, dr_fifo, dr_canonical_reg, result_builder,
                )
        #
        # New type of region
        #
        else:
            #
            # New UNIQ region
            #
            if new_reg_code == SC_CODE:  # pylint: disable=compare-to-zero
                region_index = result_builder.add_sc_region()
                result_builder.add_occorc_to_region(v, region_index)
            #
            # New or already existing IR region
            #
            elif new_reg_code == IR_CODE:
                region_index = __new_or_pair_ir(
                    v, ir_lifo, ir_canonical_reg, result_builder)
            #
            # New or already existing DR region
            #
            elif new_reg_code == DR_CODE:
                region_index = __new_or_pair_dr(
                    v, dr_fifo, dr_canonical_reg, result_builder)

        reg_code = new_reg_code
        u = v
        v = __succ_in_path(v, mdcg, var, initial_vertex)

    # ------------------------------------------------------------------------ #
    # To view
    # ------------------------------------------------------------------------ #
    return result_builder.view(prob, ilp_codes)


# pylint: disable=too-many-arguments
def __previous_or_new_ir(new_reg_code: RegionCodeT, region_index: IndexT,
                         u: OccOrCT, v: OccOrCT, var: PuLPVarModelT,
                         ir_lifo: dict[IndexT, LifoQueue[OccOrCT]],
                         ir_canonical_reg: dict[OccOrCT, IndexT],
                         result_builder: _ScaffoldingResultBuilder) -> IndexT:
    canonical_invf = invf_canonical(v)
    #
    # Contiguous or new forward IR region
    #
    if canonical_invf not in ir_canonical_reg:
        #
        # Contiguous forward IR region
        #
        if __is_repeat_contiguous(u, v, var, new_reg_code):
            result_builder.add_occorc_to_region(v, region_index)
            ir_lifo[region_index].put(invf_other(v))
            ir_canonical_reg[canonical_invf] = region_index
            return region_index
        #
        # New forward IR region
        #
        region_index = result_builder.add_ir_region()
        result_builder.add_occorc_to_region(v, region_index)
        ir_lifo[region_index] = LifoQueue()
        ir_lifo[region_index].put(invf_other(v))
        ir_canonical_reg[canonical_invf] = region_index
        return region_index
    #
    # IR region already existing
    #
    if region_index != ir_canonical_reg[canonical_invf]:
        #
        # New reverse IR region
        #
        region_index = ir_canonical_reg[canonical_invf]
        result_builder.add_ir_region(region_index)
    assert v == ir_lifo[region_index].get()
    return region_index


# pylint: disable=too-many-arguments
def __previous_or_new_dr(new_reg_code: RegionCodeT, region_index: IndexT,
                         u: OccOrCT, v: OccOrCT, var: PuLPVarModelT,
                         dr_fifo: dict[IndexT, Queue[OccOrCT]],
                         dr_canonical_reg: dict[OccOrCT, IndexT],
                         result_builder: _ScaffoldingResultBuilder) -> IndexT:
    canonical_dirf = dirf_canonical(v)
    #
    # Contiguous or new first DR region
    #
    if canonical_dirf not in dr_canonical_reg:
        #
        # Contiguous first DR region
        #
        if __is_repeat_contiguous(u, v, var, new_reg_code):
            result_builder.add_occorc_to_region(v, region_index)
            dr_fifo[region_index].put(dirf_other(v))
            dr_canonical_reg[canonical_dirf] = region_index
            return region_index
        #
        # New first DR region
        #
        region_index = result_builder.add_dr_region()
        result_builder.add_occorc_to_region(v, region_index)
        dr_fifo[region_index] = Queue()
        dr_fifo[region_index].put(dirf_other(v))
        dr_canonical_reg[canonical_dirf] = region_index
        return region_index
    #
    # DR region already existing
    #
    if region_index != dr_canonical_reg[canonical_dirf]:
        #
        # New second DR region
        #
        region_index = dr_canonical_reg[canonical_dirf]
        result_builder.add_dr_region(region_index)
    assert v == dr_fifo[region_index].get()
    return region_index


def __new_or_pair_ir(v: OccOrCT, ir_lifo: dict[IndexT, LifoQueue[OccOrCT]],
                     ir_canonical_reg: dict[OccOrCT, IndexT],
                     result_builder: _ScaffoldingResultBuilder) -> IndexT:
    canonical_invf = invf_canonical(v)
    #
    # New forward IR
    #
    if canonical_invf not in ir_canonical_reg:
        region_index = result_builder.add_ir_region()
        result_builder.add_occorc_to_region(v, region_index)
        ir_lifo[region_index] = LifoQueue()
        ir_lifo[region_index].put(invf_other(v))
        ir_canonical_reg[canonical_invf] = region_index
        return region_index
    #
    # New reverse IR
    #
    region_index = ir_canonical_reg[canonical_invf]
    result_builder.add_ir_region(region_index)
    assert v == ir_lifo[region_index].get()
    return region_index


def __new_or_pair_dr(v: OccOrCT, dr_fifo: dict[IndexT, Queue[OccOrCT]],
                     dr_canonical_reg: dict[OccOrCT, IndexT],
                     result_builder: _ScaffoldingResultBuilder) -> IndexT:
    canonical_dirf = dirf_canonical(v)
    #
    # New first DR
    #
    if canonical_dirf not in dr_canonical_reg:
        region_index = result_builder.add_dr_region()
        result_builder.add_occorc_to_region(v, region_index)
        dr_fifo[region_index] = Queue()
        dr_fifo[region_index].put(dirf_other(v))
        dr_canonical_reg[canonical_dirf] = region_index
        return region_index
    #
    # New second DR
    #
    region_index = dr_canonical_reg[canonical_dirf]
    result_builder.add_dr_region(region_index)
    assert v == dr_fifo[region_index].get()
    return region_index


# ---------------------------------------------------------------------------- #
#                               Walk In The Path                               #
# ---------------------------------------------------------------------------- #
def __pred_in_path(v: OccOrCT, mdcg: MDCGraph,
                   var: PuLPVarModelT, initial_vertex: OccOrCT) -> (
        Optional[OccOrCT]):
    """Return the predecessor of vertex v in solution path.

    Stop if the predecessor is the initial vertex.

    Parameters
    ----------
    v : OccOrCT
        Vertex
    mdcg : MDCGraph
        Multiplied doubled contig graph
    var : PuLPVarModelT
        PuLP variables
    initial_vertex : OccOrCT
        Starter vertex

    Returns
    -------
    OccOrCT, optional
        The previous vertex in solution path, None if it is the
        initial vertex

    Raises
    ------
    NotACircuit
        If the path is not a circuit
    """
    for u in mdcg.multiplied_preds(v):
        if var.x[u, v].varValue > BIN_THRESHOLD:
            if u == initial_vertex:
                return None
            return u
    raise NotACircuit()


def __succ_in_path(v: OccOrCT, mdcg: MDCGraph,
                   var: PuLPVarModelT, initial_vertex: OccOrCT) -> (
        Optional[OccOrCT]):
    """Return the successor of vertex v in solution path.

    Stop if the successor is the initial vertex.

    Parameters
    ----------
    v : OccOrCT
        Vertex
    mdcg : MDCGraph
        Multiplied doubled contig graph
    var : PuLPVarModelT
        PuLP variables
    initial_vertex : OccOrCT
        Starter vertex

    Returns
    -------
    OccOrCT, optional
        The next vertex in solution path, None if it is the
        initial vertex

    Raises
    ------
    NotACircuit
        If the path is not a circuit
    """
    for w in mdcg.multiplied_succs(v):
        if var.x[v, w].varValue > BIN_THRESHOLD:
            if w == initial_vertex:
                return None
            return w
    raise NotACircuit()


# ---------------------------------------------------------------------------- #
#                                Initialisation                                #
# ---------------------------------------------------------------------------- #
def __create_set_invf_paired(var: PuLPVarModelT,
                             fix_result: Optional[ScaffoldingResult]) -> (
        set[OccOrCT]):
    """Create set of canonical of paired inverted fragments.

    Parameters
    ----------
    var : PuLPVarModelT
        PuLP variables
    fix_result : ScaffoldingResult, optional
        Scaffolding result, by default `None`

    Returns
    -------
    set of OccOrCT
        Set of canonical of paired inverted fragments
    """
    set_invf_paired: set[OccOrCT] = set()
    #
    # Add old inverted fragments pairing
    #
    if fix_result is not None:
        for region_index in fix_result.ir_regions():
            for v in fix_result.region_occorc(region_index):
                set_invf_paired.add(invf_canonical(v))
    #
    # Add new inverted fragments pairing
    #
    if isinstance(var, PuLPVarInvFModel):
        for invf_sol in var.invf_solution():
            set_invf_paired.add(invf_sol[FORWARD_INT])
    return set_invf_paired


def __create_set_dirf_paired(var: PuLPVarModelT,
                             fix_result: Optional[ScaffoldingResult]) -> (
        set[OccOrCT]):
    """Create set of canonical of paired direct fragments.

    Parameters
    ----------
    var : PuLPVarModelT
        PuLP variables
    fix_result : ScaffoldingResult, optional
        Scaffolding result, by default `None`

    Returns
    -------
    set of OccOrCT
        Set of canonical of paired direct fragments
    """
    set_dirf_paired: set[OccOrCT] = set()
    #
    # Add old direct fragments pairing
    #
    if fix_result is not None:
        for region_index in fix_result.dr_regions():
            for v in fix_result.region_occorc(region_index):
                set_dirf_paired.add(dirf_canonical(v))
    #
    # Add new direct fragments pairing
    #
    if isinstance(var, PuLPVarDirFModel):
        for dirf_sol in var.dirf_solution():
            set_dirf_paired.add(dirf_sol[FORWARD_INT])
    return set_dirf_paired


# pylint: disable=too-many-arguments
def __find_initial(mdcg: MDCGraph, starter_vertex: OccOrCT,
                   var: PuLPVarModelT,
                   set_invf_paired: set[OccOrCT],
                   set_dirf_paired: set[OccOrCT]) -> OccOrCT:
    """The first vertex of the starter's region.

    In case of circular single copy region, it returns starter_vertex.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    var : PuLPVarModelT
        PuLP variables
    set_invf_paired : set of OccOrCT
        Set of canonical of paired inverted fragments
    set_dirf_paired : set of OccOrCT
        Set of canonical of paired direct fragments

    Returns
    -------
    OccOrCT
        The first vertex of the starter's region
    """
    v: OccOrCT = starter_vertex
    u: Optional[OccOrCT] = __pred_in_path(
        starter_vertex, mdcg, var, starter_vertex)
    # pylint: disable=compare-to-zero
    while (u is not None
           and __get_region_code(u, set_invf_paired, set_dirf_paired) == 0
           ):
        v = u
        u = __pred_in_path(v, mdcg, var, starter_vertex)
    if u is None:
        return starter_vertex
    return v


# ---------------------------------------------------------------------------- #
#                               Region Management                              #
# ---------------------------------------------------------------------------- #
def __get_region_code(v: OccOrCT, set_invf_paired: set[OccOrCT],
                      set_dirf_paired: set[OccOrCT]) -> RegionCodeT:
    """Get the code of the region for the multiplied oriented contig.

    Parameters
    ----------
    v : OccOrCT
        Multiplied oriented contig
    set_invf_paired : set of OccOrCT
        Set of canonical of paired inverted fragments
    set_dirf_paired : set of OccOrCT
        Set of canonical of paired direct fragments

    Returns
    -------
    RegionCodeT
        Code of the region (0: SC; 1: IR; 2: DR)
    """
    if dirf_canonical(v) in set_dirf_paired:
        return DR_CODE
    if ((v[COCC_IND] - v[COR_IND]) % 2 == 0  # pylint: disable=compare-to-zero
            and invf_canonical(v) in set_invf_paired):
        return IR_CODE
    return SC_CODE


def __is_repeat_contiguous(u: OccOrCT, v: OccOrCT, var: PuLPVarModelT,
                           region_code: RegionCodeT) -> bool:
    """Answer if the repeat given by its code is contiguous.

    Parameters
    ----------
    u : OccOrCT
        Multiplied oriented contig
    v : OccOrCT
        Multiplied oriented contig
    var : PuLPVarModelT
        PuLP variables
    region_code : RegionCodeT
        Region's code

    Returns
    -------
    bool
        True if repeat is contiguous, else False
    """
    #
    # IR:
    #   i (= u) -> k(= v): ok, so is there l -> j?
    #
    if region_code == IR_CODE:
        return var.x[invf_other(v), invf_other(u)].varValue > BIN_THRESHOLD
    #
    # DR:
    #   i (= u) -> k(= v): ok, so is there j -> l?
    #
    if region_code == DR_CODE:
        return var.x[dirf_other(u), dirf_other(v)].varValue > BIN_THRESHOLD
    # TODO error out of code
    return False



# -*- coding=utf-8 -*-

"""Scaffolding exceptions module."""

from collections.abc import Iterable
from pathlib import Path

from pulp import LpStatus

from khloraascaf.lib import REGION_CODE_TO_ID, RegionCodeT


# ============================================================================ #
#                                    CLASSES                                   #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                               Scaffolding Error                              #
# ---------------------------------------------------------------------------- #
class ScaffoldingError(Exception):
    """Scaffolding error exception."""

    def __init__(self, outdir: Path):
        """The Initialiser."""
        super().__init__()
        self.__outdir: Path = outdir

    def __str__(self) -> str:
        """Exception message."""
        return (
            'Scaffolding fails\n'
            f'\tSee output directory: {self.__outdir}'
        )


class RepeatScaffoldingError(Exception):
    """If the repeat scaffolding solving fails during the combination."""

    def __str__(self) -> str:
        """Exception message."""
        return 'Repeat scaffolding has failed'


class SingleCopyScaffoldingError(Exception):
    """If the single copy scaffolding solving fails during the combination."""

    def __str__(self) -> str:
        """Exception message."""
        return 'Single copy scaffolding has failed'


class CombineScaffoldingError(Exception):
    """Scaffolding combination error exception."""

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return 'The scaffolding combination has failed'


# ---------------------------------------------------------------------------- #
#                                 Region Types                                 #
# ---------------------------------------------------------------------------- #
class WrongRegionCode(Exception):
    """Wrong region code exception."""

    def __init__(self, region_code: int):
        """The Initializer."""
        super().__init__()
        self.__region_code = region_code

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return f'The region code {self.__region_code} is not correct'


# ---------------------------------------------------------------------------- #
#                                    Solver                                    #
# ---------------------------------------------------------------------------- #
class WrongSolverName(Exception):
    """Wrong solver name exception."""

    def __init__(self, solver: str):
        """The Initializer."""
        super().__init__()
        self.__solver = solver

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return f'The solver name {self.__solver} is not correct'


# ---------------------------------------------------------------------------- #
#                                Unfeasible ILP                                #
# ---------------------------------------------------------------------------- #
class _UnfeasibleILP(Exception):
    """ILP problem unfeasible exception."""

    _PROBLEM_ID = 'THE PROBLEM'

    def __init__(self, status: int, ilp_codes: Iterable[RegionCodeT]):
        super().__init__()
        self.__status: int = status
        self.__ilp_codes: tuple[RegionCodeT, ...] = tuple(ilp_codes)

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return (
            f'The {self._PROBLEM_ID} problem is unfeasible:\n'
            '\t* ILP codes: '
            + '-'.join(REGION_CODE_TO_ID[code] for code in self.__ilp_codes)
            + '\n'
            f'\t* Status: {LpStatus[self.__status]}'
        )


class UnfeasibleIR(_UnfeasibleILP):
    """IR optimisation problem unfeasible exception."""

    _PROBLEM_ID = 'Find the best inverted repeats'


class UnfeasibleDR(_UnfeasibleILP):
    """DR optimisation problem unfeasible exception."""

    _PROBLEM_ID = 'Find the best direct repeats'


class UnfeasibleSC(_UnfeasibleILP):
    """SC optimisation problem unfeasible exception."""

    _PROBLEM_ID = 'Find the best single copy regions'


# ---------------------------------------------------------------------------- #
#                                 Result Error                                 #
# ---------------------------------------------------------------------------- #
class NotACircuit(Exception):
    """Not a circuit exception."""

    def __str__(self) -> str:
        """Print the exception message.

        Returns
        -------
        str
            Exception message
        """
        return 'The found path is not a circuit'



# -*- coding=utf-8 -*-

"""PuLP path model."""


from pulp import LpAffineExpression, LpProblem, lpSum
from revsymg.index_lib import FORWARD_INT, REVERSE_INT

from khloraascaf.ilp.dirf_sets import dirf_other
from khloraascaf.ilp.invf_sets import invf_other
from khloraascaf.ilp.pulp_var_db import PuLPVarPath, PuLPVarRepFModelT
from khloraascaf.multiplied_doubled_contig_graph import (
    MULT_ATTR,
    MDCGraph,
    OccOrCT,
)
from khloraascaf.result import ScaffoldingResult


# ============================================================================ #
#                                  OBJECTIVES                                  #
# ============================================================================ #
def longuest_contiguous_repeat(prob: LpProblem,
                               var: PuLPVarRepFModelT,
                               mdcg: MDCGraph):
    """Pair of the longuest contigous repeats model objective function.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarInvFModel or PuLPVarDirFModel
        PuLP variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    """
    prob += (
        lpSum(var.m[p] for p in type(var).repeat_frag_fn(mdcg))
        + lpSum(var.isadj[q] for q in type(var).adj_repeat_frag_fn(mdcg))
    )


# ============================================================================ #
#                                  CONSTRAINTS                                 #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                      Repeats Are Sub-paths In The Graph                      #
# ---------------------------------------------------------------------------- #
def pairs_in_path(prob: LpProblem,
                  var: PuLPVarRepFModelT,
                  mdcg: MDCGraph):
    """Constraint defining the link between m and i variables.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarInvFModel or PuLPVarDirFModel
        PuLP variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    """
    for i, j in type(var).repeat_frag_fn(mdcg):
        #
        # If the repeated fragment is chosen,
        #   then the first part is in the path
        #
        prob += (
            var.m[i, j] <= var.i[i]
        )
        #
        # If the repeated fragment is chosen,
        #   then the second part is in the path
        #
        prob += (
            var.m[i, j] <= var.i[j]
        )


# ---------------------------------------------------------------------------- #
#                               Repeat Structure                               #
# ---------------------------------------------------------------------------- #
#
# Definitions of relative positioning of two vertices that involves
#   in the path and are candidate of repeats
#
def alpha_definition(prob: LpProblem,
                     var: PuLPVarRepFModelT,
                     mdcg: MDCGraph, big_m: int):
    r"""Constraint defining alpha variables.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarInvFModel or PuLPVarDirFModel
        PuLP variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    big_m : int
        Big-M constant

    Notes
    -----
    It was observed (not rigorously) that forcing :math:`\alpha_{uv} = 0`
    when :math:`y_u = y_v` is faster than forcing :math:`\alpha_{uv} = 1`
    """
    for u, v in type(var).pair_repeat_alpha_fn(mdcg):
        #
        # If v is after u, then alpha_uv = 1
        #
        prob += pos(var, mdcg, v) - pos(var, mdcg, u) <= var.alpha(u, v) * big_m
        #
        # If u is after v, then alpha_uv = 0
        #
        prob += (
            pos(var, mdcg, u) - pos(var, mdcg, v)
            <= (1 - var.alpha(u, v)) * big_m
        )
        #
        # If u and v are not positioned (pos_u + pos_v = 0),
        #   then alpha_uv = 0
        #
        prob += pos(var, mdcg, u) + pos(var, mdcg, v) >= var.alpha(u, v)


#
# Forbidden relative positioning of the pairs of repeated fragments
#
def forbidden_pairing_definition(prob: LpProblem,
                                 var: PuLPVarRepFModelT,
                                 mdcg: MDCGraph):
    """Constraint defining the forbidden pairing cases.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarInvFModel or PuLPVarDirFModel
        PuLP variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    """
    for p_repf in type(var).pair_repeat_frag_fn(mdcg):
        for case_ind, u, v, w, x in type(var).forbidden_pairing_fn(p_repf):
            #
            #  u v w x
            # --------->
            #
            prob += (
                3 * var.pairing_ban[case_ind, p_repf[0], p_repf[1]]
                <= var.alpha(u, v)
                + var.alpha(v, w)
                + var.alpha(w, x)
            )
            prob += (
                var.alpha(u, v)
                + var.alpha(v, w)
                + var.alpha(w, x)
                <= 2 + var.pairing_ban[case_ind, p_repf[0], p_repf[1]]
            )
        #
        # Repeated fragments can be paired if they are not implied
        #   in any of the forbidden pairing cases
        #
        prob += (
            var.m[p_repf[0]] + var.m[p_repf[1]]
            <= 2 - lpSum(var.pairing_ban[nested_ind, p_repf[0], p_repf[1]]
                         for nested_ind in range(8))
        )


# ---------------------------------------------------------------------------- #
#                               Repeat Contiguity                              #
# ---------------------------------------------------------------------------- #
def adjacent_fragments(prob: LpProblem,
                       var: PuLPVarRepFModelT,
                       mdcg: MDCGraph):
    """Constraint defining the adjacency of repeated fragments.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarInvFModel or PuLPVarDirFModel
        PuLP variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    """
    for u, v in type(var).adj_repeat_frag_fn(mdcg):
        # ik lj or lj ik
        # XXX must take in account that Xik AND Xki can exist simultaneously!
        p, _ = type(var).repeat_frag_builder_fn(u)
        q, _ = type(var).repeat_frag_builder_fn(v)
        #
        # If adjacent, then the first edge is chosen
        #
        prob += var.isadj[u, v] <= var.x[u, v]
        #
        # If adjacent, then the second edge is chosen
        #
        prob += (
            var.isadj[u, v]
            <= var.x[type(var).adj_repeat_other_fn((u, v))]
        )
        #
        # Is adjacent then the first inverted fragment is chosen
        #
        prob += var.isadj[u, v] <= var.m[p]
        #
        # Is adjacent then the second inverted fragment is chosen
        #
        prob += var.isadj[u, v] <= var.m[q]


# ---------------------------------------------------------------------------- #
#                              Speed-up The Model                              #
# ---------------------------------------------------------------------------- #
def occurences_priority(prob: LpProblem,
                        var: PuLPVarRepFModelT,
                        mdcg: MDCGraph):
    """Constraint for occurences priority.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarInvFModel or PuLPVarDirFModel
        PuLP variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    """
    vertices = mdcg.vertices()
    for u_ind in mdcg.repeated_contigs():
        #
        # Consecutive pairs of inverted fragments priority
        #
        for p, q in type(var).pair_repeat_consecutive_fn(mdcg, u_ind):
            prob += var.m[q] <= var.m[p]
        #
        # Vertices occurrence priority
        #
        for u_occ in range(vertices.attr(u_ind, MULT_ATTR) - 1):
            prob += (
                var.i[(u_ind, FORWARD_INT, u_occ + 1)]
                + var.i[(u_ind, REVERSE_INT, u_occ + 1)]
                <= var.i[(u_ind, FORWARD_INT, u_occ)]
                + var.i[(u_ind, REVERSE_INT, u_occ)]
            )


# ---------------------------------------------------------------------------- #
#                               Keep Old Repeats                               #
# ---------------------------------------------------------------------------- #
def fix_repeats_subpaths(prob: LpProblem, var: PuLPVarPath,
                         fix_result: ScaffoldingResult):
    """Fix subpaths corresponding to repeats.

    Parameters
    ----------
    prob : LpProblem
        PuLP problem
    var : PuLPVarPresScoreModel
        PuLP variables
    fix_result : ScaffoldingResult
        Previous scaffolding result
    """
    # ------------------------------------------------------------------------ #
    # Inverted repeats
    # ------------------------------------------------------------------------ #
    for ir_ind in fix_result.ir_regions():

        ir_path_iter = fix_result.region_occorc(ir_ind)
        i = next(ir_path_iter)
        j = invf_other(i)
        #
        # The two vertices of the beginner inverted fragment are in the path
        #
        prob += var.i[i] == 1
        prob += var.i[j] == 1

        for k in ir_path_iter:

            l = invf_other(k)
            #
            # The two vertices of the next inverted fragment are in the path
            #
            prob += var.i[k] == 1
            prob += var.i[l] == 1
            #
            # Fix edges of inverted fragments
            #
            prob += var.x[i, k] == 1
            prob += var.x[l, j] == 1

            i = k
            j = l

    # ------------------------------------------------------------------------ #
    # Direct repeats
    # ------------------------------------------------------------------------ #
    for dr_ind in fix_result.dr_regions():

        dr_path_iter = fix_result.region_occorc(dr_ind)
        i = next(dr_path_iter)
        j = dirf_other(i)
        #
        # The two vertices of the beginner direct fragment are in the path
        #
        prob += var.i[i] == 1
        prob += var.i[j] == 1

        for k in dr_path_iter:

            l = dirf_other(k)
            #
            # The two vertices of the next direct fragment are in the path
            #
            prob += var.i[k] == 1
            prob += var.i[l] == 1
            #
            # Fix edges of direct fragments
            #
            prob += var.x[i, k] == 1
            prob += var.x[j, l] == 1

            i = k
            j = l


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
def pos(var: PuLPVarRepFModelT, mdcg: MDCGraph,
        vertex: OccOrCT) -> LpAffineExpression:
    """Give the position of vertex in the path.

    Parameters
    ----------
    var : PuLPVarInvFModel or PuLPVarDirFModel
        Variables
    mdcg : MDCGraph
        Multiplied doubled contig graph
    vertex : OccOrCT
        Vertex in MDCGraph

    Returns
    -------
    LpAffineExpression
        Affine expresion interpreted as the position of the vertex in the path
    """
    return lpSum(var.f[vertex, w] for w in mdcg.multiplied_succs(vertex))



# -*- coding=utf-8 -*-

"""Intersected direct fragments PuLP ILP."""


from typing import Optional

from pulp import LpMaximize, LpProblem

from khloraascaf.ilp.pulp_circuit import (
    circuit_from_the_starter_forward,
    flow_definition,
    intermediate_in_circuit,
)
from khloraascaf.ilp.pulp_repeated_fragments import (
    adjacent_fragments,
    alpha_definition,
    fix_repeats_subpaths,
    forbidden_pairing_definition,
    longuest_contiguous_repeat,
    occurences_priority,
    pairs_in_path,
)
from khloraascaf.ilp.pulp_var_db import PuLPVarDirFModel
from khloraascaf.multiplied_doubled_contig_graph import MDCGraph, OccOrCT
from khloraascaf.result import ScaffoldingResult


# ============================================================================ #
#                                  PULP MODEL                                  #
# ============================================================================ #
def intersected_dirf_model(mdcg: MDCGraph, starter_vertex: OccOrCT,
                           fix_result: Optional[ScaffoldingResult] = None) -> (
        tuple[LpProblem, PuLPVarDirFModel]):
    """Intersected direct fragments PuLP model.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    fix_result : ScaffoldingResult, optional
        Previous scaffolding result, by default `None`

    Returns
    -------
    LpProblem
        ILP problem
    PuLPVarDirFModel
        ILP variables
    """
    # ------------------------------------------------------------------------ #
    # Constants
    # ------------------------------------------------------------------------ #
    big_m = mdcg.multiplied_card() // 2

    # ------------------------------------------------------------------------ #
    # Problem
    # ------------------------------------------------------------------------ #
    prob = LpProblem(name='inters_dirf', sense=LpMaximize)

    # ------------------------------------------------------------------------ #
    # Variables
    # ------------------------------------------------------------------------ #
    var = PuLPVarDirFModel(mdcg, starter_vertex, big_m)

    # ------------------------------------------------------------------------ #
    # Objective function
    # ------------------------------------------------------------------------ #
    longuest_contiguous_repeat(prob, var, mdcg)

    # ------------------------------------------------------------------------ #
    # Constraints
    # ------------------------------------------------------------------------ #
    #
    # Path constraints
    #
    flow_definition(prob, var, big_m, mdcg)
    circuit_from_the_starter_forward(prob, var, mdcg, starter_vertex)
    intermediate_in_circuit(prob, var, mdcg, starter_vertex)
    #
    # Repeats constraints
    #
    pairs_in_path(prob, var, mdcg)
    alpha_definition(prob, var, mdcg, big_m)
    forbidden_pairing_definition(prob, var, mdcg)
    adjacent_fragments(prob, var, mdcg)
    #
    # Fix repeats sub-paths
    #
    if fix_result is not None:
        fix_repeats_subpaths(prob, var, fix_result)
        # TODO avoid pairs for fixed sub-paths
    else:
        #
        # Speed-up the model
        #
        occurences_priority(prob, var, mdcg)

    return prob, var



# -*- coding=utf-8 -*-

"""Scaffolding module."""


from collections.abc import Callable
from pathlib import Path
from typing import Optional, Type

from pulp import LpProblem, LpStatusOptimal

from khloraascaf.exceptions import (
    CombineScaffoldingError,
    RepeatScaffoldingError,
    ScaffoldingError,
    SingleCopyScaffoldingError,
    UnfeasibleDR,
    UnfeasibleIR,
    UnfeasibleSC,
    WrongRegionCode,
)
from khloraascaf.ilp.pulp_max_dirf import intersected_dirf_model
from khloraascaf.ilp.pulp_max_invf import nested_invf_model
from khloraascaf.ilp.pulp_max_presscore import best_presscore_model
from khloraascaf.ilp.pulp_to_solver import (
    fmt_solver_log_name,
    solve_pulp_problem,
)
from khloraascaf.ilp.pulp_var_db import (
    PuLPVarDirFModel,
    PuLPVarInvFModel,
    PuLPVarModelT,
)
from khloraascaf.inputs import (
    INSTANCE_NAME_DEF,
    MULT_UPB_DEF,
    OUTDEBUG_DEF,
    OUTDIR_DEF,
    PRESSCORE_UPB_DEF,
    SOLVER_CBC,
    IdCT,
    MultT,
    PresScoreT,
)
from khloraascaf.lib import (
    DR_CODE,
    IR_CODE,
    REGION_CODE_TO_ID,
    SC_CODE,
    RegionCodeT,
)
from khloraascaf.multiplied_doubled_contig_graph import (
    MDCGraph,
    MDCGraphIDContainer,
    OccOrCT,
    first_forward_occurence,
    mdcg_with_id_from_input_files,
)
from khloraascaf.outputs import (
    fmt_contigs_of_regions_filename,
    fmt_map_of_regions_filename,
    generate_output_directory,
    write_contigs_of_regions,
    write_map_of_regions,
)
from khloraascaf.result import ScaffoldingResult, path_to_regions
from khloraascaf.run_metadata import (
    bump_debug_to_yaml,
    bump_io_config_to_yaml,
    bump_solutions_to_yaml,
)
from khloraascaf.utils_debug import (
    fmt_found_repeated_fragments_filename,
    fmt_vertices_of_regions_filename,
    write_found_repeated_fragments,
    write_vertices_of_regions,
)


# DOCU constants, functions
# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# FIXME use constant args something like that
__HEIGHT_LIM = 2
__OTHER_REPEAT_CODE = {IR_CODE: DR_CODE, DR_CODE: IR_CODE}


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                        From User's Files Main Function                       #
# ---------------------------------------------------------------------------- #
# pylint: disable=too-many-arguments
def scaffolding(
        contig_attrs: Path,
        contig_links: Path,
        contig_starter: IdCT,
        multiplicity_upperbound: MultT = MULT_UPB_DEF,
        presence_score_upperbound: PresScoreT = PRESSCORE_UPB_DEF,
        solver: str = SOLVER_CBC,
        outdir: Path = OUTDIR_DEF,
        instance_name: str = INSTANCE_NAME_DEF,
        debug: bool = OUTDEBUG_DEF) -> Path:
    """Computes the scaffolding.

    Parameters
    ----------
    contig_attrs : Path
        Contigs' attributes file path
    contig_links : Path
        Contigs' links file path
    contig_starter : IdCT
        Starter contig's identifier
    multiplicity_upperbound : MultT, optional
        Multiplicities upper bound, by default `MULT_UPB_DEF`
    presence_score_upperbound : PresScoreT, optional
        Presence score upper bound, by default `PRESSCORE_UPB_DEF`
    solver : str, optional
        MILP solver to use ('cbc' or 'gurobi'), by default `SOLVER_CBC`
    outdir : Path, optional
        Output directory path, by default `OUTDIR_DEF`
    instance_name : str, optional
        Instance name, by default `INSTANCE_NAME_DEF`
    debug : bool, optional
        Output debug or not, by default False

    Returns
    -------
    Path
        Result output directory path

    Raises
    ------
    ScaffoldingError
        If the scaffolding fails

    Notes
    -----
    A uniquely named directory will be created to store the result files.
    """
    # ------------------------------------------------------------------------ #
    # Manage options
    # ------------------------------------------------------------------------ #
    # REFACTOR find less confusable name for outdir / output_dir / outdir_gen...
    outdir_gen = outdir / generate_output_directory(instance_name)
    outdir_gen.mkdir()

    bump_io_config_to_yaml(
        outdir_gen, contig_attrs, contig_links, contig_starter,
        multiplicity_upperbound, presence_score_upperbound,
        solver, outdir, instance_name, debug,
    )

    # ------------------------------------------------------------------------ #
    # Generate scaffolding data from input files
    # ------------------------------------------------------------------------ #
    # DOCU verify docstring
    mdcg, id_container = mdcg_with_id_from_input_files(
        contig_attrs, contig_links,
        multiplicity_upperbound=multiplicity_upperbound,
        presence_score_upperbound=presence_score_upperbound,
    )

    # ------------------------------------------------------------------------ #
    # Find the solutions
    # ------------------------------------------------------------------------ #
    # TODO what to do if several optimal solutions?
    #   * solution should be a list of results?
    #   * how to tell which results was selected?
    #   * metadata on the outputs (human reading)?
    try:
        solutions = combine_scaffolding_problems(
            mdcg, first_forward_occurence(
                id_container.contig_to_vertex(contig_starter)),
            solver, outdir_gen, instance_name,
            debug=debug,
        )
    except CombineScaffoldingError as exc:
        # TODO logging exception
        raise ScaffoldingError(outdir_gen) from exc

    if not solutions:
        # TODO print 'No solution found'
        return outdir

    for solution in solutions:
        # TODO print 'n solutions found'
        __solution_scaffolding(
            solution, outdir_gen, id_container, debug, instance_name)
    # TOTEST output directory returned (especially when auto-generated)
    return outdir_gen


# ---------------------------------------------------------------------------- #
#                             Combining Scaffolding                            #
# ---------------------------------------------------------------------------- #
def combine_scaffolding_problems(mdcg: MDCGraph, starter_vertex: OccOrCT,
                                 solver: str, outdir: Path, instance_name: str,
                                 debug: bool = OUTDEBUG_DEF) -> (
        tuple[ScaffoldingResult, ...]):
    """Find a priority between the optimisation problems.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    solver : str
        MILP solver to use ('cbc' or 'gurobi')
    outdir : Path
        Output directory path
    instance_name : str
        Instance's name
    debug : bool, optional
        Output debug or not, by default False

    Returns
    -------
    tuple of ScaffoldingResult
        Best ILP results

    Raises
    ------
    CombineScaffoldingError
        The scaffolding combination has failed

    Warnings
    --------
    Files in the output directory can be erased.
    """
    # Si list vide
    #   * faire IR
    #       * mettre IR solution
    #   * faire DR
    #       * si opt value
    #           * ==, rajouter
    #           * >, reinitialiser liste avec solution
    #           * <, ne pas mettre
    # Sinon, pour chaque solution
    #   * faire complement (IR <-> DR)
    #       * si opt value
    #           * ==, rajouter
    #           * >, reinitialiser liste avec solution
    #           * <, ne pas mettre
    #
    # Pour l'instant limité par __HEIGHT_LIM
    # ------------------------------------------------------------------------ #
    # Repeat scaffolding
    # ------------------------------------------------------------------------ #
    opti_results: tuple[ScaffoldingResult, ...] = ()
    height = 0
    while not height or (height < __HEIGHT_LIM and opti_results):
        try:
            opti_results = combine_repeat_scaffolding(
                opti_results, mdcg, starter_vertex,
                solver, outdir, instance_name, debug,
            )
        except RepeatScaffoldingError as exc:
            # TODO logging logging.critical(str(exc))
            raise CombineScaffoldingError() from exc
        height += 1

    # ------------------------------------------------------------------------ #
    # Single copy regions scaffolding
    # ------------------------------------------------------------------------ #
    try:
        opti_results = combine_singlecopy_scaffolding(
            opti_results, mdcg, starter_vertex,
            solver, outdir, instance_name, debug,
        )
    except SingleCopyScaffoldingError as exc:
        # TODO logging exception here
        raise CombineScaffoldingError() from exc

    return opti_results


def combine_repeat_scaffolding(opti_results: tuple[ScaffoldingResult, ...],
                               mdcg: MDCGraph, starter_vertex: OccOrCT,
                               solver: str, outdir: Path,
                               instance_name: str,
                               debug: bool) -> tuple[ScaffoldingResult, ...]:
    """Combine repeat scaffolding.

    Parameters
    ----------
    opti_results : tuple of ScaffoldingResult
        Best scaffolding results
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    solver : str
        Solver to use
    outdir : Path
        Output directory path
    instance_name : str
        Instance name
    debug : bool
        To output debug files or not

    Returns
    -------
    tuple of ScaffoldingResult
        Best scaffolding results

    Raises
    ------
    RepeatScaffoldingError
        If one of the repeat solve fails during the combination
    """
    best_opt_value = 0.0
    new_results: list[ScaffoldingResult] = []

    for opti_result in (opti_results if opti_results else (None,)):

        l_scaffolding_to_apply: list[RegionCodeT] = (
            [IR_CODE, DR_CODE] if opti_result is None
            else [__OTHER_REPEAT_CODE[opti_result.last_ilp()]]
        )

        for region_code in l_scaffolding_to_apply:
            try:
                result = scaffolding_region(
                    region_code, mdcg, starter_vertex,
                    solver, outdir, instance_name,
                    fix_result=opti_result,
                    debug=debug,
                )
            except (UnfeasibleIR, UnfeasibleDR) as exc:
                # TODO logging print(exc, file=sys.stderr)
                raise RepeatScaffoldingError() from exc

            if result.opt_value() > best_opt_value:
                best_opt_value = result.opt_value()
                new_results = [result]
            elif result.opt_value() == best_opt_value > 0:
                new_results.append(result)

    return tuple(new_results) if new_results else opti_results


def combine_singlecopy_scaffolding(opti_results: tuple[ScaffoldingResult, ...],
                                   mdcg: MDCGraph, starter_vertex: OccOrCT,
                                   solver: str, outdir: Path,
                                   instance_name: str, debug: bool) -> (
        tuple[ScaffoldingResult, ...]):
    """Combine best previous results with single copy scaffolding.

    Parameters
    ----------
    opti_results : tuple of ScaffoldingResult
        Best scaffolding results
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    solver : str
        Solver to use
    outdir : Path
        Output directory path
    instance_name : str
        Instance name
    debug : bool
        To output debug files or not

    Returns
    -------
    tuple of ScaffoldingResult
        Best scaffolding results

    Raises
    ------
    SingleCopyScaffoldingError
        One of the single copy region scaffolding has failed
    """
    best_opt_value = 0.0
    new_results = []
    for opti_result in (opti_results if opti_results else (None,)):
        try:
            result = scaffolding_region(
                SC_CODE, mdcg, starter_vertex,
                solver, outdir, instance_name,
                fix_result=opti_result,
                debug=debug,
            )
        except UnfeasibleSC as exc:
            # TODO logging here print(exc, file=sys.stderr)
            raise SingleCopyScaffoldingError() from exc

        if result.opt_value() > best_opt_value:
            best_opt_value = result.opt_value()
            new_results = [result]
        elif result.opt_value() == best_opt_value:
            new_results.append(result)

    return tuple(new_results)


# ---------------------------------------------------------------------------- #
#                         Specific Regions Scaffolding                         #
# ---------------------------------------------------------------------------- #
def scaffolding_region(region_code: RegionCodeT,
                       mdcg: MDCGraph, starter_vertex: OccOrCT,
                       solver: str, outdir: Path, instance_name: str,
                       fix_result: Optional[ScaffoldingResult] = None,
                       debug: bool = OUTDEBUG_DEF) -> ScaffoldingResult:
    """Scaffolding of a specific region.

    Parameters
    ----------
    region_code : RegionCodeT
        Code of the region to scaffold
    mdcg : MDCGraph
        Multiplied doubled contig graph
    starter_vertex : OccOrCT
        Starter vertex
    solver : str
        MILP solver to use ('cbc' or 'gurobi')
    outdir : Path
        Output directory path
    instance_name : str
        Instance's name
    fix_result : ScaffoldingResult, optional
        Previous scaffolding result, by default None
    debug : bool, optional
        Output debug or not, by default False

    Returns
    -------
    ScaffoldingResult
        Scaffolding result

    Raises
    ------
    WrongRegionCode
        The given code of the regions is wrong
    UnfeasibleIR
        The combinatorial problem is unfeasible
    UnfeasibleDR
        The combinatorial problem is unfeasible
    UnfeasibleSC
        The combinatorial problem is unfeasible

    Warnings
    --------
    Files in the output directory can be erased.
    """
    # ------------------------------------------------------------------------ #
    # ILP codes
    # ------------------------------------------------------------------------ #
    if fix_result is not None:
        ilp_codes = tuple(fix_result.ilp_codes()) + (region_code,)
    else:
        ilp_codes = (region_code,)

    # ------------------------------------------------------------------------ #
    # Instantiate functions and exceptions
    # ------------------------------------------------------------------------ #
    model_fn: Callable[..., tuple[LpProblem, PuLPVarModelT]]
    unfeasible_exc: Type[Exception]
    if region_code == IR_CODE:
        model_fn = nested_invf_model
        unfeasible_exc = UnfeasibleIR
    elif region_code == DR_CODE:
        model_fn = intersected_dirf_model
        unfeasible_exc = UnfeasibleDR
    elif region_code == SC_CODE:
        model_fn = best_presscore_model
        unfeasible_exc = UnfeasibleSC
    else:
        raise WrongRegionCode(region_code)

    prob, var = model_fn(mdcg, starter_vertex, fix_result=fix_result)

    log_path = None
    if debug:
        log_path = outdir / fmt_solver_log_name(
            solver, instance_name,
            (REGION_CODE_TO_ID[code] for code in ilp_codes),
        )
    solve_pulp_problem(prob, solver, log_path=log_path)

    if prob.status != LpStatusOptimal:
        if debug:
            bump_debug_to_yaml(
                outdir, instance_name, ilp_codes,
                starter_vertex, prob.status, None,
            )
        raise unfeasible_exc(prob.status, ilp_codes)

    result = path_to_regions(mdcg, starter_vertex, ilp_codes,
                             prob, var, fix_result)

    # ------------------------------------------------------------------------ #
    # Debug
    # ------------------------------------------------------------------------ #
    if debug:
        __region_scaffolding_debug(instance_name, outdir, var, result)
        bump_debug_to_yaml(
            outdir, instance_name, result.ilp_codes(),
            starter_vertex, result.status(), result.opt_value(),
        )
    return result


def __region_scaffolding_debug(instance_name: str, outdir: Path,
                               var: PuLPVarModelT, result: ScaffoldingResult):
    """Write debug files and metadata.

    Parameters
    ----------
    instance_name : str
        Instance name
    outdir : Path
        Output directory path
    result : ScaffoldingResult
        Scaffolding result
    """
    # ------------------------------------------------------------------------ #
    # The order of vertices for each region
    # ------------------------------------------------------------------------ #
    vertices_of_regions_path = (
        outdir / fmt_vertices_of_regions_filename(
            instance_name,
            (REGION_CODE_TO_ID[code] for code in result.ilp_codes()),
        )
    )
    write_vertices_of_regions(result, vertices_of_regions_path)

    # ------------------------------------------------------------------------ #
    # The map of regions (debug version)
    # ------------------------------------------------------------------------ #
    map_of_regions_path = (
        outdir / fmt_map_of_regions_filename(
            instance_name,
            (REGION_CODE_TO_ID[code] for code in result.ilp_codes()),
        )
    )
    write_map_of_regions(result, map_of_regions_path)

    # ------------------------------------------------------------------------ #
    # The repeated fragments
    # ------------------------------------------------------------------------ #
    if isinstance(var, (PuLPVarInvFModel, PuLPVarDirFModel)):
        repeat_fragments_path = (
            outdir / fmt_found_repeated_fragments_filename(
                instance_name,
                (REGION_CODE_TO_ID[code] for code in result.ilp_codes()),
            )
        )
        write_found_repeated_fragments(var, repeat_fragments_path)


def __solution_scaffolding(solution: ScaffoldingResult, outdir_gen: Path,
                           id_container: MDCGraphIDContainer, debug: bool,
                           instance_name: str):
    """Output the solution in files and in YAML.

    Parameters
    ----------
    solution : ScaffoldingResult
        Scaffolding result
    outdir_gen : Path
        Generated output directory path
    id_container : MDCGraphIDContainer
        Container connecting contig identifiers and vertex indices
    debug : bool
        Debugging option
    instance_name : str
        Instance name
    """
    # -------------------------------------------------------------------- #
    # Write the solution in files
    # -------------------------------------------------------------------- #
    #
    # The order of oriented contigs for each region
    #
    contigs_of_regions_path = (
        outdir_gen / fmt_contigs_of_regions_filename(
            instance_name,
            (REGION_CODE_TO_ID[code] for code in solution.ilp_codes()),
        )
    )
    write_contigs_of_regions(
        solution, contigs_of_regions_path,
        id_container=id_container,
    )
    #
    # The order of oriented regions
    #
    if not debug:  # already generated if debug is True
        map_of_regions_path = (
            outdir_gen / fmt_map_of_regions_filename(
                instance_name,
                (REGION_CODE_TO_ID[code] for code in solution.ilp_codes()),
            )
        )
        write_map_of_regions(solution, map_of_regions_path)

    bump_solutions_to_yaml(outdir_gen, instance_name, solution.ilp_codes())
